# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2019, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------
import panel as pn

from ...utils import pretty_describe
from ..base import BaseView


class Description(BaseView):
    """
    Class for displaying a textual description of a data source.

    Parameters
    ----------
    source: intake catalog entry, or list of same
        source to describe in this object

    Attributes
    ----------
    contents: str
        string representation of the source's description
    label: str
        label to display at top of panel - contains name of source
    children: list of panel objects
        children that will be used to populate the panel when visible
    panel: panel layout object
        instance of a panel layout (row or column) that contains children
        when visible
    watchers: list of param watchers
        watchers that are set on children - cleaned up when visible
        is set to false.
    """

    main_pane = None

    def __init__(self, source=None, **kwargs):
        self.source = source
        self.panel = pn.Column(name="Description", width_policy="max", margin=0, height=240, sizing_mode="stretch_width", scroll=True)
        super().__init__(**kwargs)

    def setup(self):
        self.main_pane = pn.pane.Markdown(self.contents)
        self.children = [self.main_pane]

    @BaseView.source.setter
    def source(self, source):
        """When the source gets updated, update the pane object"""
        BaseView.source.fset(self, source)
        if self.main_pane:
            self.main_pane.object = """```yaml\n{}\n```""".format(self.contents)

    @property
    def contents(self):
        """String representation of the source's description"""
        if not self._source:
            return "name: " + "⠀" * 30
        contents = self.source.describe()
        return pretty_describe(contents)



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------
from packaging.version import Version

from intake.source.base import DataSource, Schema

from .base import RemoteSource, get_partition


class RemoteDataFrame(RemoteSource):
    """Dataframe on an Intake server"""

    name = "remote_dataframe"
    container = "dataframe"

    def __init__(self, url, headers, **kwargs):
        super(RemoteDataFrame, self).__init__(url, headers, **kwargs)
        self.npartitions = kwargs["npartitions"]
        self.shape = tuple(kwargs["shape"])
        self.metadata = kwargs["metadata"]
        self.dtype = kwargs["dtype"]
        self.verify = kwargs.get("verify", False)
        self._schema = Schema(npartitions=self.npartitions, extra_metadata=self.metadata, dtype=self.dtype, shape=self.shape)
        self.dataframe = None

    def _load_metadata(self):
        import dask.dataframe as dd
        import dask.delayed

        if self.dataframe is None:
            self.parts = [dask.delayed(get_partition)(self.url, self.headers, self._source_id, self.container, i) for i in range(self.npartitions)]
            if Version(dask.__version__) < Version("2.5.0"):
                self.dataframe = dd.from_delayed(self.parts)
            else:
                self.dataframe = dd.from_delayed(self.parts, verify_meta=self.verify)
        return self._schema

    def _get_partition(self, i):
        self._load_metadata()
        return self.parts[i].compute()

    def read(self):
        self._load_metadata()
        return self.dataframe.compute()

    def to_dask(self):
        self._load_metadata()
        return self.dataframe

    def _close(self):
        self.dataframe = None

    @staticmethod
    def _persist(source, path, **kwargs):
        """Save dataframe to local persistent store

        Makes a parquet dataset out of the data using dask.dataframe.to_parquet.
        This then becomes a data entry in the persisted datasets catalog.

        Parameters
        ----------
        source: a DataSource instance to save
        name: str or None
            Key to refer to this persisted dataset by. If not given, will
            attempt to get from the source's name
        kwargs: passed on to dask.dataframe.to_parquet
        """
        try:
            df = source.to_dask()
        except NotImplementedError:
            df = source.read()
        return RemoteDataFrame._data_to_source(df, path, **kwargs)

    @staticmethod
    def _data_to_source(df, path, **kwargs):
        import dask.dataframe as dd

        if not is_dataframe_like(df):
            raise NotImplementedError
        try:
            from intake_parquet import ParquetSource
        except ImportError:
            raise ImportError("Please install intake-parquet to use persistence" " on dataframe container sources.")
        if not hasattr(df, "npartitions"):
            df = dd.from_pandas(df, npartitions=1)
        df.to_parquet(path, **kwargs)
        source = ParquetSource(path, meta={})
        return source


def is_dataframe_like(df):
    """Looks like a Pandas DataFrame

    Copied from dask.utils
    """
    typ = type(df)
    return all(hasattr(typ, name) for name in ("groupby", "head", "merge", "mean")) and all(hasattr(df, name) for name in ("dtypes",)) and not hasattr(typ, "dtype")


class GenericDataFrame(DataSource):
    """Create partitioned dataframe from a set of files and any reader

    This data-source allows you to specify any reader to create dataframes.
    The reader must take an open file-like object as input, and output a
    pandas.DataFrame.

    Parameters
    ----------
    urlpath: str
        Location of data. May be local files or remote with a protocol
        specifier. May be a list of files or a glob pattern to be expanded.
    reader: func
        f(open_file, **kwargs) -> pandas.DataFrame
    storage_options: dict
        Passed to the file-system backend to open files; typically includes
        credentials for remote storage
    kwargs:
        Passed to reader function
    """

    name = "generic_dataframe"
    container = "dataframe"

    def __init__(self, urlpath, reader, storage_options=None, **kwargs):
        self.url = urlpath
        self.reader = reader
        self.storage_options = storage_options or {}
        kwargs = kwargs.copy()
        super().__init__(metadata=kwargs.pop("metadata", {}))
        self.kwargs = kwargs
        self.dataframe = None

    def _load_metadata(self):
        import dask.dataframe as dd
        import dask.delayed
        from fsspec import open_files

        self.files = open_files(self.url, **self.storage_options)

        def read_a_file(open_file, reader, kwargs):
            with open_file as of:
                df = reader(of, **kwargs)
                df["path"] = open_file.path
                return df

        if self.dataframe is None:
            self.parts = [dask.delayed(read_a_file)(open_file, self.reader, self.kwargs) for open_file in self.files]
            self.dataframe = dd.from_delayed(self.parts)
            self.npartitions = self.dataframe.npartitions
            self.shape = (None, len(self.dataframe.columns))
            self.dtype = self.dataframe.dtypes.to_dict()
            self._schema = Schema(npartitions=self.npartitions, extra_metadata=self.metadata, dtype=self.dtype, shape=self.shape)
        return self._schema

    def _get_partition(self, i):
        self._load_metadata()
        return self.parts[i].compute()

    def read(self):
        self._load_metadata()
        return self.dataframe.compute()

    def to_dask(self):
        self._load_metadata()
        return self.dataframe

    def _close(self):
        self.dataframe = None



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

import json
import os
import subprocess
import sys

import appdirs

from intake.config import conf
from intake.utils import make_path_posix

from .local import Catalog, YAMLFilesCatalog


def load_user_catalog():
    """Return a catalog for the platform-specific user Intake directory"""
    cat_dir = user_data_dir()
    if not os.path.isdir(cat_dir):
        return Catalog()
    else:
        return YAMLFilesCatalog(cat_dir)


def user_data_dir():
    """Return the user Intake catalog directory"""
    return appdirs.user_data_dir(appname="intake", appauthor="intake")


def load_global_catalog():
    """Return a catalog for the environment-specific Intake directory"""
    cat_dir = global_data_dir()
    if not os.path.isdir(cat_dir):
        return Catalog()
    else:
        return YAMLFilesCatalog(cat_dir)


CONDA_VAR = "CONDA_PREFIX"
VIRTUALENV_VAR = "VIRTUAL_ENV"


def conda_prefix():
    """Fallback: ask conda in PATH for its prefix"""
    try:
        out = subprocess.check_output(["conda", "info", "--json"])
        return json.loads(out.decode())["default_prefix"]
    except (subprocess.CalledProcessError, json.JSONDecodeError, OSError):
        return False


def which(program):
    """Emulate posix ``which``"""
    import distutils.spawn

    return distutils.spawn.find_executable(program)


def global_data_dir():
    """Return the global Intake catalog dir for the current environment"""
    prefix = False
    if VIRTUALENV_VAR in os.environ:
        prefix = os.environ[VIRTUALENV_VAR]
    elif CONDA_VAR in os.environ:
        prefix = sys.prefix
    elif which("conda"):
        # conda exists but is not activated
        prefix = conda_prefix()

    if prefix:
        # conda and virtualenv use Linux-style directory pattern
        return make_path_posix(os.path.join(prefix, "share", "intake"))
    else:
        return appdirs.site_data_dir(appname="intake", appauthor="intake")


def load_combo_catalog():
    """Load a union of the user and global catalogs for convenience"""
    user_dir = user_data_dir()
    global_dir = global_data_dir()
    desc = "Generated from data packages found on your intake search path"
    cat_dirs = []
    if os.path.isdir(user_dir):
        cat_dirs.append(user_dir + "/*.yaml")
        cat_dirs.append(user_dir + "/*.yml")
    if os.path.isdir(global_dir):
        cat_dirs.append(global_dir + "/*.yaml")
        cat_dirs.append(global_dir + "/*.yml")
    for path_dir in conf.get("catalog_path", []):
        if path_dir != "":
            if not path_dir.endswith(("yaml", "yml")):
                cat_dirs.append(path_dir + "/*.yaml")
                cat_dirs.append(path_dir + "/*.yml")
            else:
                cat_dirs.append(path_dir)

    return YAMLFilesCatalog(cat_dirs, name="builtin", description=desc)



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

import gzip
import io
import pickle
from collections import OrderedDict

import msgpack

from ..compat import pack_kwargs


class NoneCompressor(object):
    name = "none"

    def compress(self, data):
        return data

    def decompress(self, data):
        return data


class GzipCompressor(object):
    name = "gzip"

    def compress(self, data):
        buf = io.BytesIO()
        with gzip.GzipFile(fileobj=buf, mode="wb", compresslevel=1) as f:
            f.write(data)
        return buf.getvalue()

    def decompress(self, data):
        with gzip.GzipFile(fileobj=io.BytesIO(data)) as f:
            return f.read()


try:
    import msgpack_numpy
except ImportError:
    msgpack_numpy = None


def check_pyarrow():
    try:
        import pyarrow
    except ImportError:
        raise ImportError("Serializing DataFrames requires pyarrow.")
    return pyarrow


class MsgPackSerializer(object):
    # TODO: This is ugly, should maybe transition to
    #  distributed.protocol.serialize
    name = "msgpack"

    def encode(self, obj, container):
        if container in ["ndarray", "xarray"] and msgpack_numpy:
            from ..compat import np_pack_kwargs

            return msgpack.packb(obj, **np_pack_kwargs)
        elif container == "dataframe":
            # Use pyarrow for serializing DataFrames, rather than
            # msgpack: https://github.com/intake/intake/issues/460
            pa = check_pyarrow()

            context = pa.default_serialization_context()
            # This eventually goes to msgpack.packb, which doesn't
            # directly accept PyArrow Buffer objects. Need to wrap
            # it in a memoryview to avoid a TypeError.
            return memoryview(context.serialize(obj).to_buffer())
        else:
            return msgpack.packb(obj, **pack_kwargs)

    def decode(self, bytestr, container):
        from ..compat import unpack_kwargs

        if container in ["ndarray", "xarray"] and msgpack_numpy:
            from ..compat import np_unpack_kwargs

            return msgpack.unpackb(bytestr, **np_unpack_kwargs)
        elif container == "dataframe":
            pa = check_pyarrow()
            context = pa.default_serialization_context()
            return context.deserialize(bytestr)
        else:
            return msgpack.unpackb(bytestr, **unpack_kwargs)


class PickleSerializer(object):
    def __init__(self, protocol_level):
        self._protocol_level = protocol_level
        self.name = "pickle%d" % protocol_level

    def encode(self, obj, container):
        return pickle.dumps(obj, protocol=self._protocol_level)

    def decode(self, bytestr, container):
        return pickle.loads(bytestr)


class ComboSerializer(object):
    def __init__(self, format_encoder, compressor):
        self._format_encoder = format_encoder
        self._compressor = compressor
        self.format_name = format_encoder.name
        self.compressor_name = compressor.name

    def encode(self, obj, container):
        return self._compressor.compress(self._format_encoder.encode(obj, container))

    def decode(self, bytestr, container):
        return self._format_encoder.decode(self._compressor.decompress(bytestr), container)


compressors = [GzipCompressor(), NoneCompressor()]
try:
    import snappy

    class SnappyCompressor(object):
        name = "snappy"

        def compress(self, data):
            return snappy.compress(data)

        def decompress(self, data):
            return snappy.decompress(data)

    compressors.insert(0, SnappyCompressor())
except ImportError:
    pass


# Insert in preference order
picklers = [PickleSerializer(protocol) for protocol in [2, 1]]
serializers = [MsgPackSerializer()] + picklers
format_registry = OrderedDict([(e.name, e) for e in serializers])
compression_registry = OrderedDict([(e.name, e) for e in compressors])



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

import collections
import copy
import keyword
import logging
import posixpath
import re
import warnings

import msgpack

from intake.auth.base import AuthenticationFailure, BaseClientAuth

from ..compat import pack_kwargs, unpack_kwargs
from ..source import registry as plugin_registry
from ..utils import remake_instance
from . import Catalog
from .entry import CatalogEntry
from .utils import RemoteCatalogError, coerce, expand_defaults

logger = logging.getLogger("intake")


class RemoteCatalog(Catalog):
    """The state of a remote Intake server"""

    name = "intake_remote"

    def __init__(
        self,
        url,
        http_args=None,
        page_size=None,
        name=None,
        source_id=None,
        metadata=None,
        auth=None,
        ttl=1,
        getenv=True,
        getshell=True,
        storage_options=None,
        parameters=None,
        persist_mode="default",
    ):
        """Connect to remote Intake Server as a catalog

        Parameters
        ----------
        url: str
            Address of the server, e.g., "intake://localhost:5000".
        http_args: dict
            Arguments to add to HTTP calls, including "ssl" (True/False) for
            secure connections.
        page_size : int, optional
            The number of entries fetched at a time during iteration.
            Default is None (no pagination; fetch all entries in bulk).
        name : str, optional
            Unique identifier for catalog. This is primarily useful when
            manually constructing a catalog. Defaults to None.
        source_id : str, optional
            Emphemeral unique ID generated by the server, if known.
        metadata: dict
            Additional information about this data
        auth : BaseClientAuth or None
            Default, None, falls back to BaseClientAuth.
        ttl : float, optional
            Lifespan (time to live) of cached modification time. Units are in
            seconds. Defaults to 1.
        getenv: bool
            Can parameter default fields take values from the environment
        getshell: bool
            Can parameter default fields run shell commands
        storage_options : dict
            If using a URL beginning with 'intake://' (remote Intake server),
            parameters to pass to requests when issuing http commands; otherwise
            parameters to pass to remote backend file-system. Ignored for
            normal local files.
        parameters: dict
            To pass to the server when it instantiates the data source
        """
        from requests.compat import urljoin, urlparse

        if http_args is None:
            http_args = {}
        else:
            # Make a deep copy to avoid mutating input.
            http_args = copy.deepcopy(http_args)
        secure = http_args.pop("ssl", False)
        scheme = "https" if secure else "http"
        url = url.replace("intake", scheme, 1)
        if not url.endswith("/"):
            url = url + "/"
        self.url = url
        self.info_url = urljoin(url, "v1/info")
        self.source_url = urljoin(url, "v1/source")
        self.http_args = http_args
        self.http_args.update(storage_options or {})
        self.http_args["headers"] = self.http_args.get("headers", {})
        self._page_size = page_size
        self._source_id = source_id
        self._parameters = parameters
        self._len = None
        self.auth = auth or BaseClientAuth()

        if self._source_id is None:
            name = urlparse(url).netloc.replace(".", "_").replace(":", "_")
        super(RemoteCatalog, self).__init__(
            name=name, metadata=name, ttl=ttl, getenv=getenv, getshell=getshell, storage_options=storage_options, persist_mode=persist_mode
        )

    def _make_entries_container(self):
        return Entries(self)

    def __dir__(self):
        # Include (cached) tab-completable entries and normal attributes.
        return [
            key for key in self._ipython_key_completions_() if re.match("[_A-Za-z][_a-zA-Z0-9]*$", key) and not keyword.iskeyword(key)  # valid Python identifier
        ] + list(  # not a Python keyword
            self.__dict__.keys()
        )

    def _ipython_key_completions_(self):
        if not self._entries.complete:
            # Ensure that at least one page of data has been loaded so that
            # *some* entries are included.
            next(iter(self))
        if not self._entries.complete:
            warnings.warn("Tab-complete and dir() on RemoteCatalog may include only a " "subset of the available entries.")
        # Loop through the cached entries, but do not trigger iteration over
        # the full set.
        # Intentionally access _entries directly to avoid paying for a reload.
        return [key for key, _ in self._entries.cached_items()]

    @property
    def page_size(self):
        return self._page_size

    def fetch_page(self, page_offset):
        import requests

        logger.debug("Request page entries %d-%d", page_offset, page_offset + self._page_size)
        params = {"page_offset": page_offset, "page_size": self._page_size}
        http_args = self._get_http_args(params)
        response = requests.get(self.info_url, **http_args)
        # Produce a chained exception with both the underlying HTTPError
        # and our own more direct context.
        try:
            response.raise_for_status()
        except requests.HTTPError as err:
            raise RemoteCatalogError("Failed to fetch page of entries {}-{}." "".format(page_offset, page_offset + self._page_size)) from err
        info = msgpack.unpackb(response.content, **unpack_kwargs)
        page = {}
        for source in info["sources"]:
            # user_parameters = source.get('user_parameters', [])
            # TODO Do something with self._parameters.
            page[source["name"]] = RemoteCatalogEntry(
                url=self.url,
                getenv=self.getenv,
                getshell=self.getshell,
                auth=self.auth,
                http_args=self.http_args,
                page_size=self._page_size,
                persist_mode=self.pmode,
                # user_parameters=user_parameters,
                **source
            )
        return page

    def fetch_by_name(self, name):
        import requests

        logger.debug("Requesting info about entry named '%s'", name)
        params = {"name": name}
        http_args = self._get_http_args(params)
        response = requests.get(self.source_url, **http_args)
        if response.status_code == 404:
            raise KeyError(name)
        try:
            response.raise_for_status()
        except requests.HTTPError as err:
            raise RemoteCatalogError("Failed to fetch entry {!r}.".format(name)) from err
        info = msgpack.unpackb(response.content, **unpack_kwargs)
        return RemoteCatalogEntry(
            url=self.url,
            getenv=self.getenv,
            getshell=self.getshell,
            auth=self.auth,
            http_args=self.http_args,
            page_size=self._page_size,
            persist_mode=self.pmode,
            **info["source"]
        )

    def _get_http_args(self, params):
        """
        Return a copy of the http_args

        Adds auth headers and 'source-id', merges in params.
        """
        # Add the auth headers to any other headers
        headers = self.http_args.get("headers", {})
        if self.auth is not None:
            auth_headers = self.auth.get_headers()
            headers.update(auth_headers)

        # build new http args with these headers
        http_args = self.http_args.copy()
        if self._source_id is not None:
            headers["source-id"] = self._source_id
        http_args["headers"] = headers

        # Merge in any params specified by the caller.
        merged_params = http_args.get("params", {})
        merged_params.update(params)
        http_args["params"] = merged_params
        return http_args

    def _load(self):
        """Fetch metadata from remote. Entries are fetched lazily."""
        # This will not immediately fetch any sources (entries). It will lazily
        # fetch sources from the server in paginated blocks when this Catalog
        # is iterated over. It will fetch specific sources when they are
        # accessed in this Catalog via __getitem__.
        import requests

        if self.page_size is None:
            # Fetch all source info.
            params = {}
        else:
            # Just fetch the metadata now; fetch source info later in pages.
            params = {"page_offset": 0, "page_size": 0}
        http_args = self._get_http_args(params)
        response = requests.get(self.info_url, **http_args)
        try:
            response.raise_for_status()
            error = False
        except requests.HTTPError as err:
            if "403" in err.args[0]:
                error = "Your current level of authentication does not have access"
            else:
                raise RemoteCatalogError("Failed to fetch metadata.") from err
        if error:
            raise AuthenticationFailure(error)
        info = msgpack.unpackb(response.content, **unpack_kwargs)
        self.metadata = info["metadata"]
        # The intake server now always provides a length, but the server may be
        # running an older version of intake.
        self._len = info.get("length")
        self._entries.reset()
        # If we are paginating (page_size is not None) and the server we are
        # working with is new enough to support pagination, info['sources']
        # should be empty. If either of those things is not true,
        # info['sources'] will contain all the entries and we should cache them
        # now.
        if info["sources"]:
            # Signal that we are not paginating, even if we were asked to.
            self._page_size = None
            self._entries._page_cache.update(
                {
                    source["name"]: RemoteCatalogEntry(url=self.url, getenv=self.getenv, getshell=self.getshell, auth=self.auth, http_args=self.http_args, **source)
                    for source in info["sources"]
                }
            )

    def search(self, *args, **kwargs):
        import requests

        request = {"action": "search", "query": (args, kwargs), "source_id": self._source_id}
        response = requests.post(url=self.source_url, **self._get_http_args({}), data=msgpack.packb(request, **pack_kwargs))
        try:
            response.raise_for_status()
        except requests.HTTPError as err:
            raise RemoteCatalogError("Failed search query.") from err
        source = msgpack.unpackb(response.content, **unpack_kwargs)
        source_id = source["source_id"]
        cat = RemoteCatalog(url=self.url, http_args=self.http_args, source_id=source_id, persist_mode=self.pmode, name="")
        cat.cat = self
        return cat

    def __len__(self):
        if self._len is None:
            # The server is running an old version of intake and did not
            # provide a length, so we have no choice but to do this the
            # expensive way.
            return sum(1 for _ in self)
        else:
            return self._len

    @staticmethod
    def _persist(source, path, **kwargs):
        return RemoteCatalog._data_to_source(source, path, **kwargs)

    @staticmethod
    def _data_to_source(cat, path, **kwargs):
        import yaml
        from fsspec import open_files

        from intake.catalog.local import YAMLFileCatalog

        if not isinstance(cat, Catalog):
            raise NotImplementedError
        out = {}
        # reach down into the private state because we apparently need the
        # Entry here rather than the public facing DataSource objects.
        for name, entry in cat._entries.items():
            out[name] = entry.__getstate__()
            out[name]["parameters"] = [up._captured_init_kwargs for up in entry._user_parameters]
            out[name]["kwargs"].pop("parameters")
        fn = posixpath.join(path, "cat.yaml")
        with open_files([fn], "wt")[0] as f:
            yaml.dump({"sources": out}, f)
        return YAMLFileCatalog(fn)


class Entries(collections.abc.Mapping):
    """Fetches entries from server on item lookup and iteration.

    This fetches pages of entries from the server during iteration and
    caches them. On __getitem__ it fetches the specific entry from the
    server.
    """

    # This has PY3-style lazy methods (keys, values, items). Since it's
    # internal we should not need the PY2-only iter* variants.
    def __init__(self, catalog):
        self._catalog = catalog
        self._page_cache = collections.OrderedDict()
        # Put lookups that were due to __getitem__ in a separate cache
        # so that iteration reflects the server's order, not an
        # arbitrary cache order.
        self._direct_lookup_cache = {}
        self._page_offset = 0
        # True if all pages are cached locally
        self.complete = self._catalog.page_size is None

    def reset(self):
        "Clear caches to force a reload."
        self._page_cache.clear()
        self._direct_lookup_cache.clear()
        self._page_offset = 0
        self.complete = self._catalog.page_size is None

    def __iter__(self):
        for key in self._page_cache:
            yield key
        if self._catalog.page_size is None:
            # We are not paginating, either because the user set page_size=None
            # or the server is a version of intake before pagination parameters
            # were added.
            return
        # Fetch more entries from the server.
        while True:
            page = self._catalog.fetch_page(self._page_offset)
            self._page_cache.update(page)
            self._page_offset += len(page)
            for key in page:
                yield key
            if len(page) < self._catalog.page_size:
                # Partial or empty page.
                # We are done until the next call to items(), when we
                # will resume at the offset where we left off.
                self.complete = True
                break

    def cached_items(self):
        """
        Iterate over items that are already cached. Perform no requests.
        """
        for item in self._page_cache.items():
            yield item
        for item in self._direct_lookup_cache.items():
            yield item

    def __getitem__(self, key):
        try:
            return self._direct_lookup_cache[key]
        except KeyError:
            try:
                return self._page_cache[key]
            except KeyError:
                source = self._catalog.fetch_by_name(key)
                self._direct_lookup_cache[key] = source
                return source

    def __len__(self):
        return len(self._catalog)


class RemoteCatalogEntry(CatalogEntry):
    """An entry referring to a remote data definition"""

    def __init__(
        self,
        url,
        auth,
        name=None,
        user_parameters=None,
        container=None,
        description="",
        metadata=None,
        http_args=None,
        page_size=None,
        persist_mode="default",
        direct_access=False,
        getenv=True,
        getshell=True,
        **kwargs
    ):
        """

        Parameters
        ----------
        url: str
            HTTP address of the Intake server this entry comes from
        auth: Auth instance
            If there are additional headers to add to calls, this instance will
            provide them
        kwargs: additional keys describing the entry, name, description,
            container,
        """
        self.url = url
        if isinstance(auth, dict):
            auth = remake_instance(auth)
        self.auth = auth
        self.container = container
        self.name = name
        self.description = description
        self._metadata = metadata or {}
        self._page_size = page_size
        # Persist mode describing a nested RemoteCatalog
        self.catalog_pmode = persist_mode
        self._user_parameters = [remake_instance(up) if (isinstance(up, dict) and "cls" in up) else up for up in user_parameters or []]
        self._direct_access = direct_access
        self.http_args = (http_args or {}).copy()
        if "headers" not in self.http_args:
            self.http_args["headers"] = {}

        super(RemoteCatalogEntry, self).__init__(getenv=getenv, getshell=getshell)

        # Persist mode for the RemoteCatalogEntry
        self._pmode = "never"

    def describe(self):
        return {
            "name": self.name,
            "container": self.container,
            "plugin": "remote",
            "description": self.description,
            "direct_access": self._direct_access,
            "metadata": self._metadata,
            "user_parameters": self._user_parameters,
            "args": (self.url,),
        }

    def get(self, **user_parameters):
        for par in self._user_parameters:
            if par["name"] not in user_parameters:
                default = par["default"]
                if isinstance(default, str):
                    default = coerce(par["type"], expand_defaults(par["default"], True, self.getenv, self.getshell))
                user_parameters[par["name"]] = default

        http_args = self.http_args.copy()
        http_args["headers"] = self.http_args["headers"].copy()
        http_args["headers"].update(self.auth.get_headers())
        return open_remote(
            self.url,
            self.name,
            container=self.container,
            user_parameters=user_parameters,
            description=self.description,
            http_args=http_args,
            page_size=self._page_size,
            auth=self.auth,
            getenv=self.getenv,
            persist_mode=self.catalog_pmode,
            getshell=self.getshell,
        )


def open_remote(url, entry, container, user_parameters, description, http_args, page_size=None, persist_mode=None, auth=None, getenv=None, getshell=None):
    """Create either local direct data source or remote streamed source"""
    import msgpack
    import requests
    from requests.compat import urljoin

    from intake.container import container_map

    if url.startswith("intake://"):
        url = url[len("intake://") :]
    payload = dict(action="open", name=entry, parameters=user_parameters, available_plugins=list(plugin_registry))
    req = requests.post(urljoin(url, "v1/source"), data=msgpack.packb(payload, **pack_kwargs), **http_args)
    if req.ok:
        response = msgpack.unpackb(req.content, **unpack_kwargs)

        if "plugin" in response:
            pl = response["plugin"]
            pl = [pl] if isinstance(pl, str) else pl
            # Direct access
            for p in pl:
                if p in plugin_registry:
                    source = plugin_registry[p](**response["args"])
                    proxy = False
                    break
            else:
                proxy = True
        else:
            proxy = True
        if proxy:
            response.pop("container")
            response.update({"name": entry, "parameters": user_parameters})
            if container == "catalog":
                response.update(
                    {
                        "auth": auth,
                        "getenv": getenv,
                        "getshell": getshell,
                        "page_size": page_size,
                        "persist_mode": persist_mode
                        # TODO ttl?
                        # TODO storage_options?
                    }
                )
            source = container_map[container](url, http_args, **response)
        source.description = description
        return source
    else:
        raise Exception("Server error: %d, %s" % (req.status_code, req.reason))



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------
"""

"""

import logging

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

# Standard library imports
import os

# External imports
import yaml

# Intake imports
from intake.cli.util import Subcommand

# -----------------------------------------------------------------------------
# API
# -----------------------------------------------------------------------------


class Config(Subcommand):
    """Configuration functions"""

    name = "config"

    def initialize(self):
        sub_parser = self.parser.add_subparsers()

        list = sub_parser.add_parser("list-defaults", help="Show all builtin defaults")
        list.set_defaults(invoke=self._list_defaults)

        conf_reset = sub_parser.add_parser("reset", help="Set config file to defaults")
        conf_reset.set_defaults(invoke=self._reset)

        conf_info = sub_parser.add_parser("info", help="Show config settings")
        conf_info.set_defaults(invoke=self._info)

        conf_get = sub_parser.add_parser("get", help="Get current config, specific key or all")
        conf_get.add_argument("key", type=str, help="Key in config dictionary", nargs="?")
        conf_get.set_defaults(invoke=self._get)

    def invoke(self, args):
        self.parser.print_help()

    def _get(self, args):
        from intake.config import conf

        if args.key:
            print(conf[args.key])
        else:
            print(yaml.dump(conf, default_flow_style=False))

    def _info(self, args):
        from intake.config import cfile

        if "INTAKE_CONF_DIR" in os.environ:
            print("INTAKE_CONF_DIR: ", os.environ["INTAKE_CONF_DIR"])
        if "INTAKE_CONF_FILE" in os.environ:
            print("INTAKE_CONF_FILE: ", os.environ["INTAKE_CONF_FILE"])
        ex = "" if os.path.isfile(cfile()) else "(does not exist)"
        print("Using: ", cfile(), ex)

    def _list_defaults(self, args):
        from intake.config import defaults

        print(yaml.dump(defaults, default_flow_style=False))

    def _reset(self, args):
        from intake.config import conf

        conf.reset()
        conf.save()



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------
"""
CLI for listing, enabling, disabling intake drivers
"""

import logging

from intake.cli.util import Subcommand
from intake.source.discovery import drivers

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------
# API
# -----------------------------------------------------------------------------


class Drivers(Subcommand):
    """
    List, enable, and disable intake drivers.
    """

    name = "drivers"

    def initialize(self):
        sub_parser = self.parser.add_subparsers()

        list = sub_parser.add_parser("list", help="Show all intake drivers, whether enabled, disabled, " "or directly inserted into the registry")
        list.add_argument("-v", "--verbose", action="store_true", help="Show module path.")
        list.set_defaults(invoke=self._list)

        enable = sub_parser.add_parser("enable", help="Enable an intake driver.")
        enable.add_argument("name", type=str, help="Driver name")
        enable.add_argument("driver", type=str, default=None, nargs="?", help="Module path and class name, as in " "package.submodule.ClassName")
        enable.set_defaults(invoke=self._enable)

        disable = sub_parser.add_parser("disable", help="Disable one or more intake drivers.")
        disable.add_argument("names", type=str, help="Driver names", nargs="+")
        disable.set_defaults(invoke=self._disable)

    def invoke(self, args):
        self.parser.print_help()

    def _list(self, args):
        if drivers.do_scan:
            print("Package scan:")
            for k, v in drivers.scanned.items():
                print(f"{k:<30}{v.__module__}.{v.__name__}")
            print()

        print("Entrypoints:")
        eps = [ep for ep in drivers.from_entrypoints() if ep.name not in drivers.disabled()]
        if eps:
            for v in eps:
                print(f"{v.name:<30}{v.module_name}:{v.object_name}")
        else:
            print("<none>")
        print()

        print("From Config:")
        eps = [ep for ep in drivers.from_conf() if ep.name not in drivers.disabled()]
        if eps:
            for v in eps:
                if v.name not in drivers.disabled():
                    print(f"{v.name:<30}{v.module_name}:{v.object_name}")
        else:
            print("<none>")
        print()

        print("Disabled: ", drivers.disabled() or "<none>")

    def _enable(self, args):
        drivers.enable(args.name, args.driver)

    def _disable(self, args):
        for name in args.names:
            drivers.disable(name)



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

import logging
from collections.abc import MappingView

import entrypoints

from .base import DataSource
from .discovery import drivers

logger = logging.getLogger("intake")


class DriverRegistry(MappingView):
    """Dict of driver: DataSource class

    If the value object is a EntryPoint or str, will load it when accesses, which
    does the import.
    """

    def __init__(self, drivers_source=drivers):
        self.drivers = drivers_source

    def __getitem__(self, item):
        it = self.drivers.enabled_plugins()[item]
        if isinstance(it, entrypoints.EntryPoint):
            return it.load()
        elif isinstance(it, str):
            return import_name(it)
        elif issubclass(it, DataSource):
            return it
        raise ValueError

    def __iter__(self):
        return iter(self.drivers.enabled_plugins())

    def keys(self):
        return list(self)

    def __len__(self):
        return len(self.drivers.enabled_plugins())

    def __repr__(self):
        return """<Intake driver registry>"""

    def __contains__(self, item):
        return item in self.keys()


registry = DriverRegistry()
register_driver = drivers.register_driver
unregister_driver = drivers.unregister_driver

# A set of fully-qualified package.module.Class mappings
classes = {}


def import_name(name):
    import importlib

    if ":" in name:
        if name.count(":") > 1:
            raise ValueError("Cannot decipher name to import: %s", name)
        mod, rest = name.split(":")
        bit = importlib.import_module(mod)
        for part in rest.split("."):
            bit = getattr(bit, part)
        return bit
    else:
        mod, cls = name.rsplit(".", 1)
        module = importlib.import_module(mod)
        return getattr(module, cls)


def get_plugin_class(name):
    if name in registry:
        return registry[name]
    if "." not in name:
        logger.debug('Plugin name "%s" not known' % name)
        return None
    if name not in classes:
        try:
            classes[name] = import_name(name)
        except (KeyError, NameError, ImportError):
            logger.debug('Failed to import "%s"' % name)
    return classes.get(name, None)



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

import os
import posixpath
import subprocess
import tempfile
import time

import pytest
import requests

from intake import config, open_catalog, register_driver
from intake.container import persist
from intake.source.base import DataSource, Schema
from intake.tests.test_utils import copy_test_file
from intake.util_tests import PY2, ex
from intake.utils import make_path_posix

here = os.path.dirname(__file__)


MIN_PORT = 7480
MAX_PORT = 7489
PORT = MIN_PORT


class TestSource(DataSource):
    name = "test"
    container = "python"

    def __init__(self, **kwargs):
        self.test_kwargs = kwargs
        super().__init__()

    def _get_schema(self):
        return Schema()


register_driver("test", TestSource)


@pytest.fixture
def tmp_config_path(tmp_path):
    key = "INTAKE_CONF_FILE"
    original = os.getenv(key)
    temp_config_path = os.path.join(tmp_path, "test_config.yml")
    os.environ[key] = temp_config_path
    assert config.cfile() == temp_config_path
    yield temp_config_path
    config.conf.reset()
    if original:
        os.environ[key] = original
    else:
        del os.environ[key]
    assert config.cfile() != temp_config_path


def ping_server(url, swallow_exception, head=None):
    try:
        r = requests.get(url)
    except Exception as e:
        if swallow_exception:
            return False
        else:
            raise e

    return r.status_code in (200, 403)  # allow forbidden as well


def pick_port():
    global PORT
    port = PORT
    if port == MAX_PORT:
        PORT = MIN_PORT
    else:
        PORT += 1

    return port


@pytest.fixture(scope="module")
def intake_server(request):
    persist.PersistStore().clear()
    os.environ["INTAKE_DEBUG"] = "true"
    # Catalog path comes from the test module
    path = request.module.TEST_CATALOG_PATH
    if isinstance(path, list):
        catalog_path = [p + "/*" for p in path]
    elif isinstance(path, str) and not path.endswith(".yml") and not path.endswith(".yaml"):
        catalog_path = path + "/*"
    else:
        catalog_path = path
    server_conf = getattr(request.module, "TEST_SERVER_CONF", None)

    # Start a catalog server on nonstandard port

    env = dict(os.environ)
    env["INTAKE_TEST"] = "server"
    if server_conf is not None:
        env["INTAKE_CONF_FILE"] = server_conf
    port = pick_port()
    cmd = [ex, "-m", "intake.cli.server", "--sys-exit-on-sigterm", "--port", str(port), "--ttl", "1"]
    if isinstance(catalog_path, list):
        cmd.extend(catalog_path)
    else:
        cmd.append(catalog_path)
    try:
        p = subprocess.Popen(cmd, env=env)
        url = "http://localhost:%d/v1/info" % port

        # wait for server to finish initalizing, but let the exception through
        # on last retry
        retries = 300
        try:
            while not ping_server(url, swallow_exception=(retries > 1)):
                time.sleep(0.1)
                retries -= 1
        except Exception:
            print(p.communicate())
            raise
        assert retries > 0, "Server never appeared"

        yield "intake://localhost:%d" % port
    finally:
        if server_conf:
            try:
                env.pop("INTAKE_CONF_FILE", None)
                os.remove(server_conf)
            except Exception:
                pass
        p.terminate()
        time.sleep(1)
        p.kill()


@pytest.fixture(scope="module")
def http_server():
    port_as_str = str(pick_port())
    if PY2:
        cmd = ["python", "-m", "SimpleHTTPServer", port_as_str]
    else:
        cmd = ["python", "-m", "http.server", port_as_str]
    p = subprocess.Popen(cmd, cwd=os.path.join(here, "catalog", "tests"))
    url = "http://localhost:{}/".format(port_as_str)
    timeout = 5
    while True:
        try:
            requests.get(url)
            break
        except Exception:
            time.sleep(0.1)
            timeout -= 0.1
            assert timeout > 0, "timeout waiting for http server"
    try:
        yield url
    finally:
        p.terminate()
        p.communicate()


@pytest.fixture(scope="function")
def tempdir():
    import shutil
    import tempfile

    d = make_path_posix(str(tempfile.mkdtemp()))
    try:
        yield d
    finally:
        shutil.rmtree(d)


@pytest.fixture(scope="function")
def temp_cache(tempdir):
    import intake
    from intake.container.persist import store

    old = intake.config.conf.copy()
    olddir = intake.config.confdir
    intake.config.confdir = tempdir
    intake.config.conf.update({"cache_dir": make_path_posix(str(tempdir)), "cache_download_progress": False, "cache_disabled": False})
    intake.config.conf.save()
    store.__init__(os.path.join(tempdir, "persist"))
    try:
        yield
    finally:
        intake.config.confdir = olddir
        intake.config.conf.update(old)
        intake.config.save_conf()


@pytest.fixture(scope="function")
def env(temp_cache, tempdir):
    import intake

    env = os.environ.copy()
    env["INTAKE_CONF_DIR"] = intake.config.confdir
    env["INTAKE_CACHE_DIR"] = intake.config.conf["cache_dir"]
    return env


@pytest.fixture
def inherit_params_cat():
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = posixpath.join(tmp_dir, "intake")
        target_catalog = copy_test_file("catalog_inherit_params.yml", tmp_path)
        return open_catalog(target_catalog)


@pytest.fixture
def inherit_params_multiple_cats():
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = posixpath.join(tmp_dir, "intake")
        copy_test_file("catalog_inherit_params.yml", tmp_path)
        copy_test_file("catalog_nested_sub.yml", tmp_path)
        return open_catalog(tmp_path + "/*.yml")


@pytest.fixture
def inherit_params_subcat():
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = posixpath.join(tmp_dir, "intake")
        target_catalog = copy_test_file("catalog_inherit_params.yml", tmp_path)
        copy_test_file("catalog_nested_sub.yml", tmp_path)
        return open_catalog(target_catalog)



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

import logging
import os
import posixpath
from os.path import expanduser

import yaml

from intake.utils import yaml_load

from .utils import make_path_posix

logger = logging.getLogger("intake")

confdir = make_path_posix(os.getenv("INTAKE_CONF_DIR", os.path.join(expanduser("~"), ".intake")))


defaults = {
    "auth": {"cls": "intake.auth.base.BaseAuth"},
    "port": 5000,
    "cache_dir": posixpath.join(confdir, "cache"),
    "cache_disabled": False,
    "cache_download_progress": True,
    "logging": "INFO",
    "catalog_path": [],
    "persist_path": posixpath.join(confdir, "persisted"),
    "package_scan": False,
}


def cfile():
    return make_path_posix(os.getenv("INTAKE_CONF_FILE", posixpath.join(confdir, "conf.yaml")))


class Config(dict):
    def __init__(self, filename=None, **kwargs):
        self.filename = filename if filename is not None else cfile()
        self.reload_all()
        super().__init__(**kwargs)

    def reset(self):
        """Set conf values back to defaults"""
        self.clear()
        self.update(defaults)

    def save(self):
        """Save current configuration to file as YAML

        Uses ``.filename`` for target location
        """
        if self.filename is False:
            return
        try:
            os.makedirs(os.path.dirname(self.filename))
        except (OSError, IOError):
            pass
        with open(self.filename, "w") as f:
            yaml.dump(dict(self), f)

    def reload_all(self):
        self.reset()
        self.load()
        self.load_env()

    def load(self, fn=None):
        """Update global config from YAML file

        If fn is None, looks in global config directory, which is either defined
        by the INTAKE_CONF_DIR env-var or is ~/.intake/ .
        """
        fn = fn or self.filename

        if os.path.isfile(fn):
            with open(fn) as f:
                try:
                    self.update(yaml_load(f))
                except Exception as e:
                    logger.warning('Failure to load config file "{fn}": {e}' "".format(fn=fn, e=e))

    def load_env(self):
        """Analyse environment variables and update conf accordingly"""
        # environment variables take precedence over conf file
        for key, envvar in [["cache_dir", "INTAKE_CACHE_DIR"], ["catalog_path", "INTAKE_PATH"], ["persist_path", "INTAKE_PERSIST_PATH"]]:
            if envvar in os.environ:
                self[key] = make_path_posix(os.environ[envvar])
        self["catalog_path"] = intake_path_dirs(self["catalog_path"])
        for key, envvar in [["cache_disabled", "INTAKE_DISABLE_CACHING"], ["cache_download_progress", "INTAKE_CACHE_PROGRESS"]]:
            if envvar in os.environ:
                self[key] = os.environ[envvar].lower() in ["true", "t", "y", "yes"]
        if "INTAKE_LOG_LEVEL" in os.environ:
            self["logging"] = os.environ["INTAKE_LOG_LEVEL"]


def intake_path_dirs(path):
    """Return a list of directories from the intake path.

    If a string, perhaps taken from an environment variable, then the
    list of paths will be split on the character ":" for posix of ";" for
    windows. Protocol indicators ("protocol://") will be ignored.
    """
    if isinstance(path, (list, tuple)):
        return path
    import re

    pattern = re.compile(";" if os.name == "nt" else r"(?<!:):(?![:/])")
    return pattern.split(path)


conf = Config()
conf.reload_all()
save_conf = conf.save
load_cond = conf.load

logger.setLevel(conf["logging"].upper())
ch = logging.StreamHandler()
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - " "%(filename)s:%(funcName)s:L%(lineno)d - " "%(message)s")
ch.setFormatter(formatter)
logger.addHandler(ch)
logger.debug("Intake logger set to debug")



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

import importlib
import logging
import os
import re
import warnings

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

from .catalog.base import Catalog
from .source import registry

imports = {
    "DataSource": "intake.source.base:DataSource",
    "Schema": "intake.source.base:Schema",
    "load_combo_catalog": "intake.catalog.default:load_combo_catalog",
    "upload": "intake.container:upload",
    "gui": "intake.interface:instance",
    "interface": "intake.interface",
    "cat": "intake.catalog:builtin",
    "output_notebook": "intake.interface:output_notebook",
    "register_driver": "intake.source:register_driver",
    "unregister_driver": "intake.source:unregister_driver",
}
logger = logging.getLogger("intake")


def __getattr__(attr):
    """Lazy attribute propagator

    Defers inputs of functions until they are needed, according to the
    contents of the ``imports`` (submodules and classes) and ``openers``
    (functions which instantiate data sources directly)
    dicts. All keys in ``openers``
    must start with "open_", else they will be ignored.
    """
    gl = globals()

    if attr == "__all__":
        return __dir__()

    if attr in gl:
        return gl[attr]

    if attr in imports:
        dest = imports[attr]
        modname = dest.split(":", 1)[0]
        logger.debug("Importing: %s" % modname)
        mod = importlib.import_module(modname)
        if ":" in dest:
            mod = getattr(mod, dest.split(":")[1])
        gl[attr] = mod
        return mod

    if attr[:5] == "open_" and attr[5:] in registry.drivers.enabled_plugins():
        driver = registry[attr[5:]]  # "open_..."
        return driver

    raise AttributeError(attr)


def __dir__(*_, **__):
    openers = ["open_" + name for name in registry.drivers.enabled_plugins()]
    return sorted(list(globals()) + list(imports) + openers)


def open_catalog(uri=None, **kwargs):
    """Create a Catalog object

    Can load YAML catalog files, connect to an intake server, or create any
    arbitrary Catalog subclass instance. In the general case, the user should
    supply ``driver=`` with a value from the plugins registry which has a
    container type of catalog. File locations can generally be remote, if
    specifying a URL protocol.

    The default behaviour if not specifying the driver is as follows:

    - if ``uri`` is a single string ending in "yml" or "yaml", open it as a
      catalog file
    - if ``uri`` is a list of strings, a string containing a glob character
      ("*") or a string not ending in "y(a)ml", open as a set of catalog
      files. In the latter case, assume it is a directory.
    - if ``uri`` begins with protocol ``"intake:"``, connect to a remote
      Intake server
    - if ``uri`` is ``None`` or missing, create a base Catalog object without entries.

    Parameters
    ----------
    uri: str or pathlib.Path
        Designator for the location of the catalog.
    kwargs:
        passed to subclass instance, see documentation of the individual
        catalog classes. For example, ``yaml_files_cat`` (when specifying
        multiple uris or a glob string) takes the additional
        parameter ``flatten=True|False``, specifying whether all data sources
        are merged in a single namespace, or each file becomes
        a sub-catalog.

    See also
    --------
    intake.open_yaml_files_cat, intake.open_yaml_file_cat,
    intake.open_intake_remote
    """
    driver = kwargs.pop("driver", None)
    if isinstance(uri, os.PathLike):
        uri = os.fspath(uri)
    if driver is None:
        if uri:
            if (isinstance(uri, str) and "*" in uri) or ((isinstance(uri, (list, tuple))) and len(uri) > 1):
                # glob string or list of files/globs
                driver = "yaml_files_cat"
            elif isinstance(uri, (list, tuple)) and len(uri) == 1:
                uri = uri[0]
                if "*" in uri[0]:
                    # single glob string in a list
                    driver = "yaml_files_cat"
                else:
                    # single filename in a list
                    driver = "yaml_file_cat"
            elif isinstance(uri, str):
                # single URL
                if uri.startswith("intake:"):
                    # server
                    driver = "intake_remote"
                else:
                    if uri.endswith((".yml", ".yaml")):
                        driver = "yaml_file_cat"
                    else:
                        uri = uri.rstrip("/") + "/*.y*ml"
                        driver = "yaml_files_cat"
            else:
                raise ValueError("URI not understood: %s" % uri)
        else:
            # empty cat
            driver = "catalog"
    if "_file" not in driver:
        kwargs.pop("fs", None)
    if driver not in registry:
        raise ValueError("Unknown catalog driver (%s), supply one of: %s" % (driver, list(sorted(registry))))
    return registry[driver](uri, **kwargs)



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

import collections
import json
import logging
import os
import posixpath
import re
import shutil
import warnings
from datetime import datetime
from hashlib import md5

from intake.config import conf
from intake.utils import is_notebook, make_path_posix

logger = logging.getLogger("intake")


def sanitize_path(path):
    """Utility for cleaning up paths."""
    from fsspec.utils import infer_storage_options

    storage_option = infer_storage_options(path)

    protocol = storage_option["protocol"]
    if protocol in ("http", "https"):
        # Most FSs remove the protocol but not HTTPFS. We need to strip
        # it to match properly.
        path = os.path.normpath(path.replace("{}://".format(protocol), ""))
    elif protocol == "file":
        # Remove trailing slashes from file paths.
        path = os.path.normpath(path)
        # Remove colons
        path = path.replace(":", "")
    # Otherwise we just make sure that path is posix
    return make_path_posix(path)


display = set()


class BaseCache(object):
    """
    Provides utilities for managing cached data files.

    Providers of caching functionality should derive from this, and appear
    as entries in ``registry``. The principle methods to override are
    ``_make_files()`` and ``_load()`` and ``_from_metadata()``.
    """

    # download block size in bytes
    blocksize = 5000000

    def __init__(self, driver, spec, catdir=None, cache_dir=None, storage_options={}):
        """
        Parameters
        ----------
        driver: str
            Name of the plugin that can load catalog entry
        spec: dict
            Specification for caching the data source.
        cache_dir: str or None
            Explicit location of cache root directory
        catdir: str or None
            Directory containing the catalog from which this spec was made
        """
        self._driver = driver
        self._spec = spec
        cd = make_path_posix(cache_dir or conf["cache_dir"])
        if cd == "catdir":
            if catdir is None:
                raise TypeError('cache_dir="catdir" only allowed when loaded' "from a catalog file.")
            cd = posixpath.join(catdir, "intake_cache")
        self._cache_dir = cd

        self._storage_options = storage_options
        self._metadata = CacheMetadata()

    def _ensure_cache_dir(self):
        if not os.path.exists(self._cache_dir):
            os.makedirs(self._cache_dir)
        if os.path.isfile(self._cache_dir):
            raise Exception("Path for cache directory exists as a file: {}" "".format(self._cache_dir))

    def _munge_path(self, cache_subdir, urlpath):
        import re

        path = sanitize_path(urlpath)

        if "regex" in self._spec:
            regex = r"%s" % sanitize_path(self._spec["regex"])
            path = re.sub(regex, "", path)

        return posixpath.join(self._cache_dir, cache_subdir, path.lstrip("/\\"))

    def _hash(self, urlpath):
        return md5(str((os.path.basename(urlpath), self._spec.get("regex", ""), self._driver)).encode()).hexdigest()

    def _path(self, urlpath, subdir=None):
        if subdir is None:
            subdir = self._hash(urlpath)
        cache_path = self._munge_path(subdir, urlpath)

        dirname = os.path.dirname(cache_path)
        if not os.path.exists(dirname):
            if not (dirname.startswith("https://") or dirname.startswith("http://")):
                os.makedirs(dirname)

        return cache_path

    def _log_metadata(self, urlpath, original_path, cache_path):
        metadata = {"created": datetime.now().isoformat(), "original_path": original_path, "cache_path": cache_path}
        self._metadata.update(urlpath, metadata)

    def load(self, urlpath, output=None, **kwargs):
        """
        Downloads data from a given url, generates a hashed filename,
        logs metadata, and caches it locally.

        Parameters
        ----------

        urlpath: str, location of data
            May be a local path, or remote path if including a protocol specifier
            such as ``'s3://'``. May include glob wildcards.
        output: bool
            Whether to show progress bars; turn off for testing

        Returns
        -------
        List of local cache_paths to be opened instead of the remote file(s). If
        caching is disable, the urlpath is returned.
        """
        if conf.get("cache_disabled", False):
            return [urlpath]
        self.output = output if output is not None else conf.get("cache_download_progress", True)

        cache_paths = self._from_metadata(urlpath)
        if cache_paths is None or any(not os.path.exists(c) for c in cache_paths):
            files_in, files_out = self._make_files(urlpath)
            self._load(files_in, files_out, urlpath)
        cache_paths = self._from_metadata(urlpath)
        return cache_paths

    def _from_metadata(self, urlpath):
        """Return set of local URLs if files already exist"""
        md = self.get_metadata(urlpath)
        if md is not None:
            return [e["cache_path"] for e in md]

    def _load(self, files_in, files_out, urlpath, meta=True):
        """Download a set of files"""
        import dask

        out = []
        outnames = []
        for file_in, file_out in zip(files_in, files_out):
            cache_path = file_out.path
            outnames.append(cache_path)

            # If `_munge_path` did not find a match we want to avoid
            # writing to the urlpath.
            if cache_path == urlpath:
                continue

            if not os.path.isfile(cache_path):
                logger.debug("Caching file: {}".format(file_in.path))
                logger.debug("Original path: {}".format(urlpath))
                logger.debug("Cached at: {}".format(cache_path))
                if meta:
                    self._log_metadata(urlpath, file_in.path, cache_path)
                ddown = dask.delayed(_download)
                out.append(ddown(file_in, file_out, self.blocksize, self.output))
        dask.compute(*out)
        return outnames

    def _make_files(self, urlpath, **kwargs):
        """Make OpenFiles for all input/outputs"""
        raise NotImplementedError

    def get_metadata(self, urlpath):
        """
        Parameters
        ----------

        urlpath: str, location of data
            May be a local path, or remote path if including a protocol specifier
            such as ``'s3://'``. May include glob wildcards.

        Returns
        -------
        Metadata (dict) about a given urlpath.
        """
        return self._metadata.get(urlpath)

    def clear_cache(self, urlpath):
        """
        Clears cache and metadata for a given urlpath.

        Parameters
        ----------

        urlpath: str, location of data
            May be a local path, or remote path if including a protocol specifier
            such as ``'s3://'``. May include glob wildcards.
        """
        cache_entries = self._metadata.pop(urlpath, [])  # ignore if missing
        for cache_entry in cache_entries:
            try:
                os.remove(cache_entry["cache_path"])
            except (OSError, IOError):
                pass
        try:
            fn = os.path.dirname(cache_entry["cache_path"])
            os.rmdir(fn)
        except (OSError, IOError):
            logger.debug("Failed to remove cache directory: %s" % fn)

    def clear_all(self):
        """
        Clears all cache and metadata.
        """
        for urlpath in self._metadata.keys():
            self.clear_cache(urlpath)

        # Safely clean up anything else.
        if not os.path.isdir(self._cache_dir):
            return
        for subdir in os.listdir(self._cache_dir):
            try:
                fn = posixpath.join(self._cache_dir, subdir)
                if os.path.isdir(fn):
                    shutil.rmtree(fn)
                if os.path.isfile(fn):
                    os.remove(fn)
            except (OSError, IOError) as e:
                logger.warning(str(e))


def _download(file_in, file_out, blocksize, output=False):
    """Read from input and write to output file in blocks"""
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        if output:
            try:
                from tqdm.autonotebook import tqdm
            except ImportError:
                logger.warn("Cache progress bar requires tqdm to be installed:" " conda/pip install tqdm")
                output = False

        if is_notebook():
            try:
                import ipywidgets  # noqa
            except ImportError:
                logger.warn("Cache progress bar in a notebook requires ipywidgets to be installed:" " conda/pip install ipywidgets")
                output = False

        try:
            file_size = file_in.fs.size(file_in.path)
            pbar_disabled = not file_size
        except ValueError as err:
            logger.debug("File system error requesting size: {}".format(err))
            file_size = 0
            pbar_disabled = True
        if output:
            if not pbar_disabled:
                for i in range(100):
                    if i not in display:
                        display.add(i)
                        out = i
                        break
                pbar = tqdm(
                    total=file_size // 2**20,
                    leave=False,
                    disable=pbar_disabled,
                    position=out,
                    desc=os.path.basename(file_out.path),
                    mininterval=0.1,
                    bar_format=r"{n}/|/{l_bar}",
                )
            else:
                output = False

        logger.debug("Caching {}".format(file_in.path))
        with file_in as f1:
            with file_out as f2:
                data = True
                while data:
                    data = f1.read(blocksize if file_size else -1)
                    f2.write(data)
                    if output:
                        pbar.update(len(data) // 2**20)
        if output:
            try:
                pbar.update(pbar.total - pbar.n)  # force to full
                pbar.close()
            except Exception as e:
                logger.debug("tqdm exception: %s" % e)
            finally:
                display.remove(out)


class FileCache(BaseCache):
    """Cache specific set of files

    Input is a single file URL, URL with glob characters or list of URLs. Output
    is a specific set of local files.
    """

    def _make_files(self, urlpath, **kwargs):
        from fsspec import open_files

        self._ensure_cache_dir()
        if isinstance(urlpath, (list, tuple)):
            subdir = self._hash(":".join(urlpath))
        else:
            subdir = self._hash(urlpath)
        files_in = open_files(urlpath, "rb", **self._storage_options)
        files_out = [open_files([self._path(f.path, subdir)], "wb", **self._storage_options)[0] for f in files_in]
        return files_in, files_out

    def _from_metadata(self, urlpath):
        return super()._from_metadata(urlpath)


class DirCache(BaseCache):
    """Cache a complete directory tree

    Input is a directory root URL, plus a ``depth`` parameter for how many
    levels of subdirectories to search. All regular files will be copied. Output
    is the resultant local directory tree.
    """

    def _make_files(self, urlpath, **kwargs):
        from fsspec import open_files

        self._ensure_cache_dir()
        subdir = self._hash(urlpath)
        depth = self._spec["depth"]
        files_in = []
        for i in range(1, depth + 1):
            files_in.extend(open_files("/".join([urlpath] + ["*"] * i)))
        files_out = [open_files([self._path(f.path, subdir)], "wb", **self._storage_options)[0] for f in files_in]
        files_in2, files_out2 = [], []
        paths = set(os.path.dirname(f.path) for f in files_in)
        for fin, fout in zip(files_in, files_out):
            if fin.path in paths:
                try:
                    os.makedirs(fout.path)
                except Exception:
                    pass
            else:
                files_in2.append(fin)
                files_out2.append(fout)
        return files_in2, files_out2

    def _from_metadata(self, urlpath):
        """Return set of local URLs if files already exist"""
        md = self.get_metadata(urlpath)
        if md is not None:
            return [self._path(urlpath)]


class CompressedCache(BaseCache):
    """Cache files extracted from downloaded compressed source

    For one or more remote compressed files, downloads to local temporary dir and
    extracts all contained files to local cache. Input is URL(s) (including
    globs) pointing to remote compressed files, plus optional ``decomp``,
    which is "infer" by default (guess from file extension) or one of the
    key strings in ``intake.source.decompress.decomp``. Optional ``regex_filter``
    parameter is used to load only the extracted files that match the pattern.
    Output is the list of extracted files.
    """

    def _make_files(self, urlpath, **kwargs):
        import tempfile

        d = tempfile.mkdtemp()
        from fsspec import open_files

        self._ensure_cache_dir()
        self._urlpath = urlpath
        files_in = open_files(urlpath, "rb", **self._storage_options)
        files_out = [open_files([make_path_posix(os.path.join(d, os.path.basename(f.path)))], "wb")[0] for f in files_in]
        super(CompressedCache, self)._load(files_in, files_out, urlpath, meta=False)
        return files_in, files_out

    def _load(self, files_in, files_out, urlpath, meta=True):
        from .decompress import decomp

        subdir = self._path(urlpath)
        try:
            os.makedirs(subdir)
        except (OSError, IOError):
            pass
        files = [f.path for f in files_out]
        out = []
        for f, orig in zip(files, files_in):
            # TODO: add snappy, brotli, lzo, lz4, xz... ?
            if "decomp" in self._spec and self._spec["decomp"] != "infer":
                d = self._spec["decomp"]
            elif f.endswith(".zip"):
                d = "zip"
            elif f.endswith(".tar.gz") or f.endswith(".tgz"):
                d = "tgz"
            elif f.endswith(".tar.bz2") or f.endswith(".tbz"):
                d = "tbz"
            elif f.endswith(".tar"):
                d = "tar"
            elif f.endswith(".gz"):
                d = "gz"
            elif f.endswith(".bz2"):
                d = "bz"
            if d not in decomp:
                raise ValueError('Unknown compression for "%s"' % f)
            out2 = decomp[d](f, subdir)
            out3 = filter(re.compile(self._spec.get("regex_filter", ".*")).search, out2)
            for fn in out3:
                logger.debug("Caching file: {}".format(f))
                logger.debug("Original path: {}".format(orig.path))
                logger.debug("Cached at: {}".format(fn))
                if meta:
                    self._log_metadata(self._urlpath, orig.path, fn)
                out.append(fn)
        return out


class DATCache(BaseCache):
    r"""Use the DAT protocol to replicate data

    For details of the protocol, see https://docs.datproject.org/
    The executable ``dat`` must be available.

    Since in this case, it is not possible to access the remote files
    directly, this cache mechanism takes no parameters. The expectation
    is that the url passed by the driver is of the form:

    ::

        dat://<dat hash>/file_pattern

    where the file pattern will typically be a glob string like "\*.json".
    """

    def _make_files(self, urlpath, **kwargs):
        self._ensure_cache_dir()
        return None, None

    def _load(self, _, __, urlpath, meta=True):
        import subprocess

        from fsspec import open_files

        path = os.path.join(self._cache_dir, self._hash(urlpath))
        dat, part = os.path.split(urlpath)
        cmd = ["dat", "clone", dat, path, "--no-watch"]
        try:
            subprocess.call(cmd, stdout=subprocess.PIPE)
        except (IOError, OSError):  # pragma: no cover
            logger.info("Calling DAT failed")
            raise
        newpath = os.path.join(path, part)

        if meta:
            for of in open_files(newpath):
                self._log_metadata(urlpath, urlpath, of.path)


class CacheMetadata(collections.abc.MutableMapping):
    """
    Utility class for managing persistent metadata stored in the Intake config directory.
    """

    def __init__(self, *args, **kwargs):
        from intake import config

        self._path = posixpath.join(make_path_posix(config.confdir), "cache_metadata.json")
        d = os.path.dirname(self._path)
        if not os.path.exists(d):
            os.makedirs(d)

        if os.path.isfile(self._path):
            with open(self._path) as f:
                self._metadata = json.load(f)
        else:
            self._metadata = {}

    def __getitem__(self, key):
        return self._metadata[self.__keytransform__(key)]

    def __setitem__(self, key, value):
        self._metadata[self.__keytransform__(key)] = value
        self._save()

    def __delitem__(self, key):
        del self._metadata[self.__keytransform__(key)]
        self._save()

    def __iter__(self):
        return iter(self._metadata)

    def __len__(self):
        return len(self._metadata)

    def __keytransform__(self, key):
        if isinstance(key, (list, tuple)):
            key = ":".join(key)
        return key

    def update(self, key, cache_entry):
        key = self.__keytransform__(key)
        entries = self._metadata.get(key, [])
        entries.append(cache_entry)
        self._metadata[key] = entries
        self._save()

    def _save(self):
        with open(self._path, "w") as f:
            json.dump(self._metadata, f)

    def pop(self, key, default=None):
        item = self._metadata.pop(key, default)
        self._save()
        return item

    def keys(self):
        return list(self._metadata.keys())


registry = {"file": FileCache, "dir": DirCache, "compressed": CompressedCache, "dat": DATCache}


def make_caches(driver, specs, catdir=None, cache_dir=None, storage_options={}):
    """
    Creates Cache objects from the cache_specs provided in the catalog yaml file

    Parameters
    ----------

    driver: str
        Name of the plugin that can load catalog entry
    specs: list
        Specification for caching the data source.
    """
    if specs is None:
        return []
    out = []
    for spec in specs:
        if "type" in spec and spec["type"] not in registry:
            raise IndexError(spec["type"])
        out.append(registry.get(spec["type"], FileCache)(driver, spec, catdir=catdir, cache_dir=cache_dir, storage_options=storage_options))
    return out



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

import os

from intake.utils import make_path_posix


def unzip(f, outpath):
    import zipfile

    z = zipfile.ZipFile(f, "r")
    z.extractall(outpath)
    out = [make_path_posix(os.path.join(outpath, fn.filename)) for fn in z.filelist]
    z.close()
    return out


def untargz(f, outpath):
    import tarfile

    tar = tarfile.open(f, "r:gz")
    out = [make_path_posix(os.path.join(outpath, fn.name)) for fn in tar.getmembers()]
    tar.extractall(outpath)
    tar.close()
    return out


def untarbz(f, outpath):
    import tarfile

    tar = tarfile.open(f, "r:bz2")
    out = [make_path_posix(os.path.join(outpath, fn.name)) for fn in tar.getmembers()]
    tar.extractall(outpath)
    tar.close()
    return out


def untar(f, outpath):
    import tarfile

    tar = tarfile.open(f, "r:")
    out = [make_path_posix(os.path.join(outpath, fn.name)) for fn in tar.getmembers()]
    tar.extractall(outpath)
    tar.close()
    return out


def ungzip(f, outpath):
    import gzip

    z = gzip.open(f)
    fn = os.path.basename(f)[:-3]
    with open(os.path.join(outpath, fn), "wb") as fout:
        data = True
        while data:
            data = z.read(2**15)
            fout.write(data)
    return [make_path_posix(os.path.join(outpath, fn))]


def unbzip(f, outpath):
    import bz2

    z = bz2.open(f)
    fn = os.path.basename(f)[:-3]
    with open(os.path.join(outpath, fn), "wb") as fout:
        data = True
        while data:
            data = z.read(2**15)
            fout.write(data)
    return [make_path_posix(os.path.join(outpath, fn))]


decomp = {"zip": unzip, "tgz": untargz, "tbz": untarbz, "tar": untar, "gz": ungzip, "bz": unbzip}
